from discminer.mining_control import _mining_channels_peakint
from discminer.mining_utils import (
    get_noise_mask,
    get_2d_plot_decorators,
    init_data_and_model,
    load_moments,
    load_disc_grid,
    show_output
)

from discminer.cube import Cube
from discminer.model import ReferenceModel
from discminer.plottools import (
    use_discminer_style,
    get_discminer_cmap,
    make_up_ax,
    get_continuous_cmap,
    make_1d_legend
)

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from astropy import units as u

import json
import subprocess

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_channels_peakint(None)
    args = parser.parse_args()

#**************************
#JSON AND SOME DEFINITIONS
#**************************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']
Rout = best['intensity']['Rout']

nchans = args.nchans
chan_step = args.step

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment)

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc

mol_tex = {
    '12co': r'$^{12}$CO',
    '13co': r'$^{13}$CO',
    'cs': r'CS'
}

try:
    mol_tex_meta = mol_tex[meta['mol']]
except KeyError:
    mol_tex_meta = meta['mol']

#*******************
#LOAD DATA AND MODEL
#*******************
Rmax = 25*u.au #Larger than sky xmax=18.25au
datacube, model = init_data_and_model(Rmin=0, Rmax=Rmax)
noise_mean, mask = get_noise_mask(datacube, thres=args.sigma)

vchannels = datacube.vchannels
pix_downsamp = model.grid['step']*meta['downsamp_fit']/au_to_m

#Some plot definitions
xmax = model.skygrid['xmax'] 
xlim = np.min([xmax/au_to_m, 25])
extent= np.array([-xmax, xmax, -xmax, xmax])/au_to_m

beam_au = datacube.beam_size.to('au').value
R_lev = np.linspace(args.Rinner, 2.096511, 4)*au_to_m

#**************************
#MAKE MODEL (2D ATTRIBUTES)
#**************************
modelcube = model.make_model(make_convolve=True)

#LOAD DISC GEOMETRY
R, phi, z = load_disc_grid()

Xproj = R[args.surface]*np.cos(phi[args.surface])/au_to_m
Yproj = R[args.surface]*np.sin(phi[args.surface])/au_to_m

r"""#Uncomment for Kelvin
datacube.data /= 1000 
modelcube.data /= 1000
modelcube.convert_to_tb(writefits=False)
datacube.convert_to_tb(writefits=False)
#"""

#*************************
#LOAD MOMENT MAPS
moment_data, moment_model, residuals, mtags = load_moments(args, mask=mask)

if args.moment=='peakintensity': #in Kelvin
    r"""#Uncomment for Kelvin
    moment_data /= 1000
    moment_model /= 1000    
    residuals /= 1000
    #"""
    pass

#*****************
#PLOT CHANNEL MAPS
#*****************
noise_mean, mask = get_noise_mask(datacube, thres=args.sigma) #again, but in Tb units
print ('Mean RMS: %.2e K'%noise_mean)

cmap = None
"""
if 'sum' in meta['kind']:
    import cmasher as cmr
    cmap = cmap_mom = plt.get_cmap('cmr.rainforest_r')
    cmap = plt.get_cmap('cmr.rainforest_r')
"""
idlim = int(0.5*chan_step*(nchans-1))
plot_channels = np.linspace(-idlim,idlim,nchans) + np.argmin(np.abs(vchannels-best['velocity']['vsys']))
plot_channels = np.append([0], plot_channels).astype(int) #dummy channel to make room for peakintensity map later on at ax[0]

kw_channels = dict(channels={'indices': plot_channels}, ncols=nchans+1,
                   xlims = (-xlim, xlim), ylims = (-xlim, xlim), show_beam=False,
                   observable = 'intensity_2',
                   #contours_from=model,
                   #kwargs_contour = dict(colors='w', linewidths=1.0), 
                   unit_coordinates='au', unit_intensity=None, projection=None)
                   
fig, ax = plt.subplots(nrows=8, ncols=nchans+1, figsize=(2.5*(nchans+1), 2.5*8)) #Master figure
plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=0.05, hspace=0.05)

#CHANNEL MAPS
levels=np.linspace(0*args.sigma*noise_mean, 4*custom['Ilim'], 32)

fig, ax0, im0, cbar0 = datacube.make_channel_maps(fig=fig, ax=ax[0,:], levels=levels, extend='max', annotate_channels=True*0, fmt_cbar='%2d', **kw_channels)
cbar0.set_label('mJy/beam')

temperatures = [25, 100, 150, 200, 400, 600, 800]
subpix = [11, 11, 11, 11, 21, 21, 21]
steps = [20000, 20000, 10000, 10000, 10000, 10000, 10000, 10000]
nT = len(temperatures)

axs = [] #Axes for spectra

nc, nx, ny = datacube.data.shape
nh = int(round(0.5*nx))
nw = 3
dy = 0.02
stat_func = np.sum #i.e. total flux 

data_slice = datacube.data[:, nh-nw:nh+nw+0, nh-nw:nh+nw+0] / datacube.beam_area #convert from mJy/beam to mJy/pixel
data_spec = np.array([stat_func(chan) for chan in data_slice])

data_max = np.max(data_spec)
data_spec/=data_max

nc, sx, sy = data_slice.shape

ax[0, 1].set_ylabel('Folded Data', fontsize=17, color='k')

rectx = model.skygrid['xygrid'][0][nh-nw:nh+nw+0]/au_to_m
xr, yr = rectx[0], rectx[0]
dxr = rectx[-1] - rectx[0]
print ('Rectangle size [au]', dxr)

for i,axi in enumerate(ax[0,1:]):
    rect = patches.Rectangle((xr, yr), dxr, dxr, 
                             facecolor='none', edgecolor='red', 
                             linewidth=1.5, alpha=0.7, fill=False)
    axi.add_patch(rect)
    axi.set_title(r'%.1f$^{\rm km/s}$'%datacube.vchannels[plot_channels[i+1]], fontsize=17)


init_funcs = {
    'velocity': model.velocity_func,
    'z_upper' : model.z_upper_func,
    'z_lower' : model.z_lower_func,
    'intensity' : model.intensity_func,
    'linewidth' : model.linewidth_func,
    'lineslope' : model.lineslope_func,
    'line_profile' : model.line_profile,
    'line_uplow' : model.line_uplow                        
}
    
F_line_arr = []

for i in range(nT):
    stepi = steps[i]
    temp = temperatures[i]
    spix = subpix[i]
    cmd_list = [
        "discminer", "parfile",
        "-f", f"log_pars_hd163296_hcn_0p1_maps_cpd_subpix{spix}_Routmin0p5_PAinit2p5_optthick_ctI0_T{temp}K_cube_60walkers_{stepi}steps.txt",
        "-o", "1"
    ]
    subprocess.run(cmd_list, check=True)
    
    dd, modeli = init_data_and_model(Rmin=0, Rmax=Rmax, subpixels=spix)
    noise_mean, mask = get_noise_mask(datacube, thres=args.sigma)
    modelcube = modeli.make_model(make_convolve=True)

    fig, axi, imi, cbari = modelcube.make_channel_maps(fig=fig, ax=ax[i+1,:], levels=im0[0].levels, annotate_channels=False, cmap=cmap, extend='max', **kw_channels)
    cbari.remove()

    with open('parfile.json') as json_file:
        pars = json.load(json_file)

    best = pars['best_fit']
    Rout = best['intensity']['Rout']
    xc = best['orientation']['xc']
    yc = best['orientation']['yc']    
    vsys_cpd = best['velocity']['vsys']
    
    for axj in ax[i+1, :]:
        modeli.make_disc_axes(axj, Rmax=1.0*Rout*u.au, color='red')
        
    ax[i+1, 1].set_ylabel('T=%dK'%temp, fontsize=17, color='k')

    pos = ax[i+1, -1].get_position()
    axsi = fig.add_axes([pos.x1+0.02, pos.y0+dy, pos.width, pos.height-2*dy])

    model_slice = modelcube.data[:, nh-nw:nh+nw+0, nh-nw:nh+nw+0] / modelcube.beam_area #convert from mJy/beam to mJy/pixel 
    model_spec = np.array([stat_func(chan) for chan in model_slice]) / data_max

    """#Plot all spectra
    for i in range(sx):
        for j in range(sy):
            axsi.step(datacube.vchannels, data_slice[:,i,j], lw=0.1, color='k', where='mid')
            axsi.step(modelcube.vchannels, model_slice[:,i,j], lw=0.2, color='magenta', where='mid')       
    """

    axsi.step(datacube.vchannels, data_spec, lw=2, color='k', where='mid', label='Data')
    axsi.fill_between(datacube.vchannels, data_spec, color='k', step='mid', alpha=0.2)
    
    axsi.step(modelcube.vchannels, model_spec, lw=2, color='goldenrod', where='mid', label='Model')
    axsi.fill_between(modelcube.vchannels, model_spec, color='yellow', step='mid', alpha=0.6)
    
    axsi.axvline(vsys_cpd, ls='--', lw=2.0, dash_capstyle='round', dashes=(2.0, 2.5), color='magenta') #, label=r'$\upsilon_{\rm CPD}$')
     
    axsi.set_ylim(-0.2, 1.2) # 2.5)    
    axsi.tick_params(labelleft=False, left=True, right=True, labelsize=16)        
    axs.append(axsi)

    r"""#Show full resolution model of the CPD velocity field
    #init_header = {}
    modelfullres = ReferenceModel(init_params=best, reset_params=True, init_funcs=init_funcs, Rmin=0, Rmax=Rmax, subpixels=0, npix=200, beam=None)

    vel2d = modelfullres.model.props[0]['upper']
    int2d = modelfullres.model.props[1]['upper'] #masked out beyond Rout
    vel2d = np.where(int2d==0.0, 0, vel2d)

    print (modelfullres.model.extent, extent, model.extent)
    ax[i+1, 0].contourf(vel2d, extent=modelfullres.model.extent, levels=np.linspace(-0.8,0.8,32)+vsys_cpd, cmap=get_discminer_cmap('velocity'))
    ax[i+1, 0].set_xlim(-1.5*Rout+xc, 1.5*Rout+xc)
    ax[i+1, 0].set_ylim(-1.5*Rout+yc, 1.5*Rout+yc)
    #if i==3:
    #    break
    #"""

    #r""" #Compute native fluxes
    modelfullres = ReferenceModel(init_params=best, reset_params=True, init_funcs=init_funcs, Rmin=0, Rmax=Rmax, subpixels=0, npix=200, beam=None)
    I = modelfullres.data
 
    # Pixel sizes (arcsec)
    pix_old = 0.039999999999996004 #I0 is in jy/pixel for this pixel size
    pix_new = modelfullres.pix_size.value*3600 #pixel size of model: 0.0025125628140702
    
    # Scale factor to fix mJy/pixel to the *actual* pixel area
    scale = (pix_new / pix_old)**2   # ~1/253
    
    I_corr = I * scale  # corrected mJy/pixel for high-res pixels
    
    # Now integrate as before
    F_nu = I_corr.sum(axis=(1,2))              # mJy per channel
    dv = np.diff(modelfullres.vchannels)
    dv = np.concatenate(([dv[0]], 0.5*(dv[:-1]+dv[1:]), [dv[-1]]))
    F_line = np.sum(F_nu * dv)                 # mJy km/s
    print(F_line, "mJy km s^-1")

    F_line_arr.append(F_line)
    #"""
    
axs[-1].set_xlabel('l.o.s velocity [km/s]', fontsize=13)
axs[-1].yaxis.set_label_position("right")
axs[-1].set_ylabel('Flux Density', fontsize=13, labelpad=25, rotation=-90)
make_1d_legend(axs[-1], ncol=2, fontsize=15, handlelength=0.8, handletextpad=0.2, columnspacing=0.5)

#DECORATIONS
"""
for axi in ax[:-1,0]:
    axi.set_xlabel('')    
    axi.tick_params(labelleft=False, labelbottom=False)

ax[0,0].set_title('Peak Intensity', color='0.6', fontsize=13, pad=8, loc='center')
ax[0,0].set_ylabel(mol_tex_meta.upper() + ' Data', color='k', fontsize=13, labelpad=6)
ax[1,0].set_ylabel('Model', color='k', fontsize=13, labelpad=4)
"""

#PLOT MOMENT MAP
for axi in ax[:,0]:
    #r"""
    for artist in axi.get_lines() + axi.collections + axi.texts:
        artist.remove()
    #"""
    axi.axis('off')

#kwargs_im = dict(cmap=cmap_mom, extent=extent, levels=im0[0].levels[::1])

#pm0 = ax[0,0].contourf(moment_data, extend='both', **kwargs_im)
#pm1 = ax[1,0].contourf(moment_model, extend='both', **kwargs_im)
#pm2 = ax[2,0].contourf(residuals, cmap=cmap_res, origin='lower', extend='both', extent=extent, levels=np.linspace(-clim, clim, 32))

#PLOT BEAM
"""
for axi in ax[0,:]:
    datacube.plot_beam(axi, fc='gold')        
"""

"""
#for axi in ax[:,int(round(0.5*(nchans+1)))-1]:
for axi in ax[:,0]:
    model.make_emission_surface(
        axi,
        R_lev=R_lev,
        kwargs_R={'colors': '0.1', 'linewidths': 0.5},
        kwargs_phi={'colors': '0.1', 'linewidths': 0.4}
    )
    model.make_disc_axes(axi, Rmax=1.2*Rout*u.au)
    for spine in ['left',  'top', 'right', 'bottom']:
        axi.spines[spine].set_color('0.8')        
    axi.tick_params(which='both', color='0.8')
    #axi.axis('off')
"""

print ('Maximum Flux Density [mJy] from data:', data_max)
print ('temperatures [K] and line fluxes [mJy km s-1]:', temperatures, np.round(F_line_arr, decimals=2))

plt.savefig('channel_maps_peakint_%s_%s_final.png'%(meta['disc'], meta['mol']), bbox_inches = 'tight', dpi=200)    
show_output(args)



from discminer.mining_control import _mining_skewkurt
from discminer.mining_utils import get_noise_mask, load_disc_grid
from discminer.core import Data

from astropy import units as u
from astropy.io import fits
import numpy as np

import os
import json
import subprocess

if __name__ == '__main__':
    parser = _mining_skewkurt(None)
    args = parser.parse_args()

#**************************
#JSON AND SOME DEFINITIONS
#**************************    
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['metadata']

dpc = meta['dpc']*u.pc
tag = meta['tag']

#**********
#LOAD DATA
#**********
file_data = 'cube_data_%s_convtb.fits'%tag
file_model = 'cube_model_%s_convtb.fits'%tag

#LOAD DISC GEOMETRY
R, phi, z = load_disc_grid()

phi_up = phi['upper']
phi_low = phi['lower']

def swap_halfplanes(a, b, mask):
    """Swap values between arrays a and b wherever mask is True."""
    a2 = a.copy()
    b2 = b.copy()
    a2[mask] = b[mask]
    b2[mask] = a[mask]
    return a2, b2

#*****************
#MAKE MOMENT MAPS
#*****************
def bm_percentiles(filepath, which='data', bluered=True, mirror=args.mirror):
    """
    Compute line percentile maps using `bettermoments`, if available.

    Parameters
    ----------
    filepath : str
        Path to the input data or model cube.

    which : {'data', 'model'}, optional
        Select whether percentiles are computed from the observed data cube
        ('data') or from a model cube ('model'). Default is 'data'.

    bluered : bool, optional
        If True, compute blue-red percentile metrics (e.g., bluewidth minus
        redwidth) instead of returning the individual components. Default is True.

    mirror : bool, optional
        If True, mirror the blue- and red-width metrics across the minor axis
        of the disc. Because emission from the backside of the disc is
        antisymmetric in velocity, the red- and blue-width maps trace opposite
        disc surfaces on either side of the minor axis and are therefore not
        directly comparable. This option enforces a consistent front/back
        assignment across the disc. Default is True.

    Notes
    -----
    This function requires the external `bettermoments` package to be installed
    and accessible from the command line.
    """

    try:
        betterm = subprocess.run(['bettermoments', filepath, '-method', 'percentiles'], capture_output=True, text=True, check=True)
        print("Calculating percentile maps with bettermoments:", betterm.stdout)
        
        for old, new in zip(['_wpdVb', '_dwpdVb', '_wpdVr', '_dwpdVr', '_wp50', '_dwp50', '_wp1684', '_dwp1684'], ['bluewidth_', 'delta_bluewidth_', 'redwidth_', 'delta_redwidth_', 'medianvelocity_', 'delta_medianvelocity_', 'meanvelocity_', 'delta_meanvelocity_']):
            os.rename(filepath.split('.fits')[0]+old+'.fits', new+'gaussian_%s.fits'%which)

        if mirror: 
            # mask for azimuths: phi > +pi/2 or phi < -pi/2 using both disc surfaces
            mirror_mask = (phi_up > np.pi/2) | (phi_up < -np.pi/2) | (phi_low > np.pi/2) | (phi_low < -np.pi/2)

            bluewidth, bluehdr = fits.getdata(f'bluewidth_gaussian_{which}.fits', header=True)
            redwidth, redhdr  = fits.getdata(f'redwidth_gaussian_{which}.fits',  header=True)
            delta_bluewidth, delta_bluehdr = fits.getdata(f'delta_bluewidth_gaussian_{which}.fits', header=True)
            delta_redwidth, delta_redhdr = fits.getdata(f'delta_redwidth_gaussian_{which}.fits', header=True)
            
            bluewidth, redwidth = swap_halfplanes(bluewidth, redwidth, mirror_mask)
            delta_bluewidth, delta_redwidth = swap_halfplanes(delta_bluewidth, delta_redwidth, mirror_mask)

            fits.writeto(f'bluewidth_gaussian_{which}.fits', bluewidth, bluehdr, overwrite=True)
            fits.writeto(f'redwidth_gaussian_{which}.fits',  redwidth,  redhdr,  overwrite=True)
            fits.writeto(f'delta_bluewidth_gaussian_{which}.fits', delta_bluewidth, delta_bluehdr, overwrite=True)
            fits.writeto(f'delta_redwidth_gaussian_{which}.fits', delta_redwidth, delta_redhdr, overwrite=True)
            
        if bluered:
            bluewidth, bluehdr = fits.getdata(f'bluewidth_gaussian_{which}.fits', header=True)
            redwidth, redhdr  = fits.getdata(f'redwidth_gaussian_{which}.fits',  header=True)
            
            fits.writeto(f'bluered_gaussian_{which}.fits', bluewidth-redwidth, header=bluehdr, overwrite=True)
            
    except FileNotFoundError as e:
        print(f"Skipping calculation of percentile maps as Bettermoments may not be installed on your system: {e}")
        print("Try installing it via 'pip install bettermoments' or upgrade it to the latest stable version using 'pip install -U bettermoments'")
    
if args.fit_data:
    datacube = Data(file_data, dpc)    
    moments_data = datacube.make_skewkurt(mtype=args.mtype, tag='data')
    bm_percentiles(file_data, which='data')
    
if args.fit_model:
    modelcube = Data(file_model, dpc)
    moments_model = modelcube.make_skewkurt(mtype=args.mtype, tag='model')
    bm_percentiles(file_model, which='model')



import numpy as np
from scipy.interpolate import LinearNDInterpolator
import scipy.sparse as sp

def get_griddata_sparse(old_coord, new_coord):
    
    #print('generating new interpolation function')
    
    xi = np.array(new_coord).T
    old_xi_shape = xi.shape
    xi = xi.reshape(-1, xi.shape[-1])

    #old coord can be an array with shape (N, ndim) or a tuple of arrays of N elements
    if isinstance(old_coord, tuple):
        old_coord = np.array(old_coord).T
    ndim = old_coord.shape[1]
        
    #construct the triangulation using LinearNDInterpolator
    interp = LinearNDInterpolator(old_coord, np.ones((old_coord.shape[0], 1)))
    simplex_indices = interp.tri.find_simplex(xi)

    #construct the interpolation matrix in COO format
    row_indices, col_indices, values = [], [], []

    #TODO: I think this can be done without the python for loop
    for n in range(xi.shape[0]):
        isimplex = simplex_indices[n]
        if isimplex == -1:
            continue  
        
        indices = interp.tri.simplices[isimplex]
        weights = [*(interp.tri.transform[isimplex, :ndim, :ndim] @ 
                     (xi[n] - interp.tri.transform[isimplex, ndim, :])), 
                   1 - (interp.tri.transform[isimplex, :ndim, :ndim] @ 
                        (xi[n] - interp.tri.transform[isimplex, ndim, :])).sum()]

        row_indices.extend([n] * len(indices))
        col_indices.extend(indices)
        values.extend(weights)

    #Convert to CSR format (more efficient for matrix multiplication)
    c = sp.coo_matrix((values, (row_indices, col_indices)), 
                      shape=(xi.shape[0], old_coord.shape[0])).tocsr()

    return lambda values: (c @ values).reshape(*old_xi_shape[:-1]).T



from discminer.mining_control import _mining_mirror_spectra
from discminer.tools.fit_kernel import _gauss
from discminer.grid import GridTools
from discminer.plottools import (MEDIUM_SIZE,
                                 make_round_map,
                                 make_up_ax,
                                 mod_major_ticks,
                                 use_discminer_style)

from discminer.mining_utils import (get_2d_plot_decorators,
                                    get_noise_mask,
                                    load_moments,
                                    load_disc_grid,
                                    init_data_and_model,
                                    make_1d_legend,
                                    show_output)

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from astropy import units as u
from scipy import interpolate
from scipy.signal import savgol_filter
from scipy.optimize import curve_fit

import json
import warnings

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_mirror_spectra(None)
    args = parser.parse_args()

integrate = 0

#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']
PA = best['orientation']['PA']
xc = best['orientation']['xc']
yc = best['orientation']['yc']
z_pars = best['height_upper']

gaps = custom['gaps']
rings = custom['rings']

try:
    kinks = custom['kinks']
except KeyError:
    kinks = []
    
ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment, unit_simple=True, fmt_vertical=True)

#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc
Rmax = 1.1*Rout*u.au #Max model radius, 10% larger than disc Rout

with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
xlim = 1.15*np.min([xmax, Rmax.value])
extent= np.array([-xmax, xmax, -xmax, xmax])

#********************
#LOAD DATA AND GRID
#********************
datacube, model = init_data_and_model()
vchannels = datacube.vchannels
model.make_model()

z_func = model.z_upper_func
vel2d, int2d, linew2d, lineb2d = model.props

R, phi, z = load_disc_grid()
noise_mean, mask = get_noise_mask(datacube, thres=2,
                                  mask_phi={'map2d': np.degrees(phi['upper']),
                                            'lims': args.mask_phi},
                                  mask_R={'map2d': R['upper']/au_to_m,
                                          'lims': args.mask_R}
)

Xproj = R[args.surface]*np.cos(phi[args.surface])
Yproj = R[args.surface]*np.sin(phi[args.surface])

X = Xproj/au_to_m
Y = Yproj/au_to_m

#**************************
#LOAD AND CLIP MOMENT MAPS
#**************************
moment_data, moment_model, residuals, mtags = load_moments(
    args,
    mask=mask,
    clip_Rmin=0.0*u.au,
    clip_Rmax=args.Router*Rout*u.au,
    clip_Rgrid=R[args.surface]*u.m
)
    
datacube.convert_to_tb(writefits=False)

#********************************
#FOLD SPECTRA TO GET CPD SIGNAL?
#********************************
R_au = R[args.surface]/au_to_m
R_au_flat = R_au.flatten()
phi_rad = phi[args.surface]
indices = np.arange(len(R_au_flat))

def _wrap_dphi(dphi):
    """Wrap angle differences to [-pi, pi] elementwise."""
    return np.arctan2(np.sin(dphi), np.cos(dphi))

def get_pixel_from_Rphi(
    Rp_au,
    phip_rad,
    tol=datacube.beam_size.to('au').value/4,
    require_within_tol=True,
):
    """
    Infer (i,j) pixel indices corresponding to disc-plane (Rp, phip),
    using the same selection strategy as get_mirror_pixel below:
      - arc-length constraint: Rp * |dphi| < tol
      - radial constraint: |R - Rp| < tol
    Then pick the best candidate by minimising hypot(dR, Rp*|dphi|).
    """
    #Flattened diffs (with 2pi wrapping)
    dphi_flat = np.abs(_wrap_dphi(phi_rad - phip_rad)).flatten()
    dR_flat = np.abs(R_au_flat - Rp_au)

    #Candidate selection
    pind = (Rp_au * dphi_flat) < tol
    rind = dR_flat[pind] < tol

    if np.any(pind) and np.any(rind):
        #Evaluate best among candidates
        Rcand = R_au_flat[pind][rind]
        dphicand = dphi_flat[pind][rind]
        totdiff = np.hypot(Rcand - Rp_au, Rp_au * dphicand)
        indbest = indices[pind][rind][np.argmin(totdiff)]
        return np.unravel_index(indbest, R_au.shape)

    #Global nearest in the same metric
    totdiff_all = np.hypot(R_au_flat - Rp_au, Rp_au * dphi_flat)
    indbest = indices[np.argmin(totdiff_all)]
    ij = np.unravel_index(indbest, R_au.shape)

    if require_within_tol:
        #Warning if diff is beyond tolerance.
        i, j = ij
        if (np.abs(R_au[i, j] - Rp_au) > tol) or (Rp_au * np.abs(_wrap_dphi(phi_rad[i, j] - phip_rad)) > tol):
            warnings.warn(
                f"Requested (Rp,phip)=({Rp_au:.2f} au, {np.degrees(phip_rad):.2f} deg) "
                f"has no candidates within tol={tol:.2f} au; using global nearest pixel {ij}."
            )

    return ij

def build_ij_window(ipix, jpix, dpix, shape):
    """Return a clipped list of (i,j) indices for a (2*dpix+1)^2 window."""
    ny, nx = shape
    i0 = max(ipix - dpix, 0)
    i1 = min(ipix + dpix, ny - 1)
    j0 = max(jpix - dpix, 0)
    j1 = min(jpix + dpix, nx - 1)

    ilist, jlist = np.meshgrid(np.arange(i0, i1 + 1), np.arange(j0, j1 + 1), indexing="ij")
    return list(zip(ilist.ravel(), jlist.ravel()))

def cli_coords_to_disc_Rphi(args, dpc, z_func, z_pars, incl, PA, xc=0.0, yc=0.0):
    """
    Returns (Rp_au, phip_rad) always in disc frame.

    - If args.input_coords == 'disc':
        rp   : au
        phip : deg (disc azimuth)
    - If args.input_coords == 'sky':
        rp   : arcsec
        phip : deg (PA from North)

    If multiple rp/phip values are passed, only the first one
    is used and a warning is printed.
    """

    rp_list = args.rp
    phip_list = args.phip

    if not isinstance(rp_list, (list, tuple, np.ndarray)):
        rp_list = [rp_list]

    if not isinstance(phip_list, (list, tuple, np.ndarray)):
        phip_list = [phip_list]

    # Print warning if more than one entry
    if len(rp_list) > 1 or len(phip_list) > 1:
        warnings.warn(
            "Multiple -rp/-phip values received. "
            "Only the first entry will be used as the reference pixel."
        )

    rp0 = float(rp_list[0])
    phip0 = float(phip_list[0])

    if args.input_coords == "disc":
        Rp_au = rp0
        phip_rad = np.deg2rad(phip0)
        return Rp_au, phip_rad

    if args.input_coords == "sky":
        rsky_au = rp0 * (dpc.to("pc").value) #arcsec to au
        PAsky = np.deg2rad(phip0)  #PA measured from North

        #Sky plane Cartesian (North up, East left)
        xsky = -rsky_au * np.sin(PAsky)
        ysky =  rsky_au * np.cos(PAsky)

        #Transform to disc frame
        xdisc, ydisc = GridTools.get_disc_from_sky_coords(
            xsky, ysky, z_func, z_pars, incl, PA, xc=xc, yc=yc
        )

        Rp_au = np.hypot(xdisc, ydisc)
        phip_rad = np.arctan2(ydisc, xdisc)

        return Rp_au, phip_rad

    raise ValueError(
        f"Unknown --input_coords '{args.input_coords}' "
        "(expected 'disc', 'disk', or 'sky')."
    )

def get_mirror_pixel(i, j, surface=args.surface, tol=datacube.beam_size.to('au').value/4):

    Rij = R_au[i,j]
    phiij = phi_rad[i,j]

    if phiij>0:
        diff = 0.5*np.pi - phiij
        phidiff = phi_rad - (0.5*np.pi + diff) #phi2d minus perfect mirror values, to be minimised along with dR
    else: 
        diff = -0.5*np.pi - phiij
        phidiff = phi_rad - (-0.5*np.pi + diff)        
        
    phidiff = np.abs(phidiff).flatten()

    pind = Rij * phidiff < tol #arc length
    rind = np.abs(Rij - R_au_flat[pind]) < tol #dR

    Rmir = R_au_flat[pind][rind]
    phimir_diff = phidiff[pind][rind]
    totdiff = np.hypot(Rmir-Rij, Rmir*phimir_diff)

    indbest = indices[pind][rind][np.argmin(totdiff)]

    return np.unravel_index(indbest, R_au.shape)

def get_mirror_channel_index(vchannels, channel_id, vsys):
    """
    Return the index of the channel mirrored around vsys.
    """
    v0 = vchannels[channel_id]
    v_mirror = 2.0 * vsys - v0

    idx_mirror = np.argmin(np.abs(vchannels - v_mirror))
    return idx_mirror

def get_spectra(cube, ij, ijmirr, vsys=vsys):

    i, j = ij
    imirr, jmirr = ijmirr

    spec = cube.data[:, i, j]
    vel_spec = vchannels - vsys
    
    spec_mirr = cube.data[:, imirr, jmirr]
    vel_mirr = vsys - vchannels

    return spec, vel_spec, spec_mirr, vel_mirr

def median_and_uncertainty(values):
    peak = np.max(values)    
    median = np.median(values)
    lower = np.percentile(values, 16)
    upper = np.percentile(values, 84)
    return peak, median, lower, upper

def fold_and_plot_spectra(cube, ijlist, vel_interp=np.linspace(-5.0, 5.0, 101), ntop=10, vsys=vsys, integrate=integrate):
    
    residuals_spec = []
    spec_arr = []
    mirr_arr = []
    
    for ij in ijlist:

        print ('Pixel, mirror:', ij, get_mirror_pixel(*ij))
        spec, vel_spec, spec_mirr, vel_mirr = get_spectra(cube, ij, get_mirror_pixel(*ij), vsys=vsys)

        if np.max(vel_interp)>np.max(vel_spec): #Fix if spectrum velocity falls out of interpolation range
            vel_interp = np.linspace(-np.max(vel_spec), np.max(vel_spec), 101)

        peakind = np.argmax(spec)
        f_spec = interpolate.interp1d(vel_spec, spec, kind='cubic')
        f_mirr = interpolate.interp1d(vel_mirr, spec_mirr, kind='cubic')

        spec_arr.append(spec)
        mirr_arr.append(spec_mirr)
        
        residuals_spec.append(f_spec(vel_interp) - f_mirr(vel_interp))

    residuals_spec = np.asarray(residuals_spec)

    #Initial guess for gaussian fits
    p0 = [0.5*np.nanmax(residuals_spec), vchannels[args.channel_id]-vsys, 0.3, 0.0] #Amplitude, centroid, sigma, offset    
    
    if integrate:
        spec_mean = np.sum(spec_arr, axis=0)
        mirr_mean = np.sum(mirr_arr, axis=0)
        #vel_spec is the same for all spectra (spec) and equally spaced
        fs = interpolate.interp1d(vel_spec, spec_mean, kind='cubic') 
        fm = interpolate.interp1d(vel_mirr, mirr_mean, kind='cubic')         
        #residuals_mean = fs(vel_interp) - fm(vel_interp)
        
    #**********************
    #PLOT RESIDUAL SPECTRA
    #**********************     
    fig, ax = plt.subplots(figsize=(14,5))

    if args.units=='Jy':
        residuals_spec*=1000
        ax.set_ylabel('Folded Residuals [mJy/beam]', fontsize=MEDIUM_SIZE+2)
    else:
        ax.set_ylabel('Folded Residuals [K]', fontsize=MEDIUM_SIZE+2)
        
    ax.set_xlabel(r'Velocity relative to CSD $\upsilon_{\rm sys}$ [km/s]', fontsize=MEDIUM_SIZE+2)
    
    for resspec in residuals_spec:
        ax.plot(vel_interp, resspec, color='k', lw=0.6, alpha=0.1)
                    
    max_curve = np.max(residuals_spec, axis=0)
    min_curve = np.min(residuals_spec, axis=0)
    
    ymax = savgol_filter(max_curve, 10, 2)      
    ymin = savgol_filter(min_curve, 10, 2)

    res_mean = np.mean(residuals_spec)
    res_std = np.std(residuals_spec)
    ax.axhline(res_mean, ls='--', lw=1.2, color='k', label='Mean (%.1f)'%res_mean)
    ax.axhline(res_mean-res_std, ls='--', lw=0.6, color='k', label=r'1$\sigma$ (%.1f)'%res_std)
    ax.axhline(res_mean+res_std, ls='--', lw=0.6, color='k')
    
    make_up_ax(ax, labelbottom=True, labeltop=False, labelsize=MEDIUM_SIZE+1, xlims=(-4.5, 4.5), ylims=(-np.max(ymax)*0.6, 1.4*np.max(ymax)))
    leg = make_1d_legend(ax, fontsize=MEDIUM_SIZE+2, handlelength=1.5, loc='lower left', bbox_to_anchor=(0.0, 1.0),
                         #frameon=True, fancybox=True, shadow=False, framealpha=1, facecolor='w', edgecolor='0.8', borderpad=1.0
    )
    #leg.get_frame().set_linewidth(3.0)
    mod_major_ticks(ax, axis='x', nbins=10)

    ax.yaxis.get_major_locator().set_params(integer=True)

    #**************************
    #GAUSSIAN FIT TO RESIDUALS
    #**************************
    amplitudes = []
    centers = []
    sigmas = []
    offsets = []
    Rlist = []
    philist = []
    rskylist = []
    PAlist = []
    
    for i, profile in enumerate(residuals_spec):
        try:
            popt, _ = curve_fit(_gauss, vel_interp, profile, p0=p0)
            A_fit, mu_fit, sigma_fit, off_fit = popt
            amplitudes.append(A_fit)
            centers.append(mu_fit)
            sigmas.append(sigma_fit)
            offsets.append(off_fit)
            ip, jp = ijlist[i]
            Rtmp = np.hypot(X[ip,jp], Y[ip,jp]) #This is correct when model parfile z=0
            phitmp = np.arctan2(Y[ip,jp], X[ip,jp])
            ztmp = model.z_upper_func({'R': Rtmp*au_to_m}, **z_pars)/au_to_m
            Rlist.append(Rtmp)
            philist.append(np.degrees(phitmp))
            xs, ys, zs = GridTools.get_sky_from_disc_coords(Rtmp, phitmp, z=ztmp, incl=incl, PA=PA, xc=xc, yc=yc)
            rskylist.append(np.hypot(xs, ys)/dpc.value) #arcsecs
            PAlist.append(270 + np.degrees(np.arctan2(ys, xs)))
            
        except RuntimeError:
            # Skip fits that fail
            continue

    amplitudes = np.array(amplitudes)
    centers = np.array(centers)
    sigmas = np.array(sigmas)
    offsets = np.array(offsets)    
    Rlist = np.array(Rlist)
    philist = np.array(philist)
    rskylist = np.array(rskylist)
    PAlist = np.array(PAlist)
    
    top_idx = np.argsort(amplitudes)[-ntop:][::-1]
    print ('R[au], phi[deg] of peak spectrum', Rlist[top_idx[0]], philist[top_idx[0]])
    print ('R[au], phi[deg] median of top %d spectra'%ntop, np.median(Rlist[top_idx]), np.median(philist[top_idx]))    
    print ('R[au], phi[deg] stddev of top %d spectra'%ntop, np.std(Rlist[top_idx]), np.std(philist[top_idx]))
    print ('Rsky[''], PA[deg] median of top %d spectra'%ntop, np.median(rskylist[top_idx]), np.median(PAlist[top_idx]))    
    print ('Rsky[''], PA[deg] stddev of top %d spectra'%ntop, np.std(rskylist[top_idx]), np.std(PAlist[top_idx])) 
    
    _, A_median, A_16, A_84 = median_and_uncertainty(amplitudes[top_idx])
    _, mu_median, mu_16, mu_84 = median_and_uncertainty(centers[top_idx])
    _, sigma_median, sigma_16, sigma_84 = median_and_uncertainty(sigmas[top_idx])
    _, off_median, off_16, off_84 = median_and_uncertainty(offsets[top_idx])    

    A_peak, mu_peak, sigma_peak, off_peak = amplitudes[top_idx[0]], centers[top_idx[0]], sigmas[top_idx[0]], offsets[top_idx[0]]
    
    print("Median Amplitude: {:.2f} (+{:.2f}/-{:.2f})".format(A_median, A_84 - A_median, A_median - A_16))
    print("Median Centroid:    {:.2f} (+{:.2f}/-{:.2f})".format(mu_median, mu_84 - mu_median, mu_median - mu_16))
    print("Median Sigma:     {:.2f} (+{:.2f}/-{:.2f})".format(sigma_median, sigma_84 - sigma_median, sigma_median - sigma_16))

    gaussian_median = _gauss(vel_interp, A_median, mu_median, sigma_median, off_median)
    vel_fit = np.linspace(vel_interp[0], vel_interp[-1], 400)

    for idx in top_idx:
        A, mu, sigma, offset = amplitudes[idx], centers[idx], sigmas[idx], offsets[idx]
        g = _gauss(vel_fit, A, mu, sigma, offset)
        ax.plot(vel_fit, g, lw=4, color='tomato', alpha=0.15, zorder=0)        
            
    print ('Peak Gaussian fit properties (A, center, sigma [FWHM], offset):', A_peak, mu_peak, sigma_peak, [2.35482*sigma_peak], off_peak)
    
    return fig, ax, residuals_spec

#*****************************************
#GET REFERENCE PIXEL, WINDOW, and MIRRORS
#*****************************************
Rp_au, phip_rad = cli_coords_to_disc_Rphi(args, dpc, z_func, z_pars, incl, PA, xc=xc, yc=yc)

ipix, jpix = get_pixel_from_Rphi(Rp_au, phip_rad) #Reference pixel
ijlist = build_ij_window(ipix, jpix, args.dpix, R_au.shape)
imir, jmir = get_mirror_pixel(ipix, jpix) #Mirror pixel

print(f"CLI coords: input-->{args.input_coords} rp={args.rp[0]} phip={args.phip[0]}")
print(f"Disc-frame target: R={Rp_au:.2f} au, phi={np.degrees(phip_rad):.2f} deg")
print(f"Reference pixel: (i,j)=({ipix},{jpix})  ->  R={R_au[ipix,jpix]:.2f} au, phi={np.degrees(phi_rad[ipix,jpix]):.2f} deg")
print(f"Mirror pixel: (i,j)=({imir},{jmir})  ->  R={R_au[imir,jmir]:.2f} au, phi={np.degrees(phi_rad[imir,jmir]):.2f} deg")

#***********************************
#SPECTRUM FROM REF PIXEL AND MIRROR
#***********************************
vchannels_fold = vchannels - vsys
        
mpl.rcParams['hatch.linewidth'] = 0.1

fig, ax = plt.subplots(figsize=(14,5))
data = datacube.data

if integrate:    
    stat_func = np.mean
    npix = args.dpix

    data_slice = data[:, ipix-npix:ipix+npix+1, jpix-npix:jpix+npix+1]
    data_spec = np.array([stat_func(chan) for chan in data_slice])

    mirr_slice = data[:, imir-npix:imir+npix+1, jmir-npix:jmir+npix+1]
    mirr_spec = np.array([stat_func(chan) for chan in mirr_slice])

    ax.step(vchannels_fold, data_spec, where='mid', lw=3.3, color='tomato')
    ax.fill_between(vchannels_fold, data_spec, step='mid', facecolor='tomato', alpha=0.1)
    
    ax.step(vchannels_fold, mirr_spec, where='mid', lw=3.3, color='dodgerblue')
    ax.fill_between(vchannels_fold, mirr_spec, step='mid', facecolor='dodgerblue', alpha=0.1)
    
    ax.step(-vchannels_fold, mirr_spec, where='mid', lw=2.2, color='k')
    ax.fill_between(-vchannels_fold, mirr_spec, step='mid', facecolor='0.5', alpha=0.6, zorder=20)
    ax.fill_between(-vchannels_fold, mirr_spec, step='mid', facecolor='none', edgecolor='k', lw=0, hatch='X', zorder=21)

else:
    ax.step(vchannels_fold, data[:, ipix, jpix], where='mid', lw=3.3, color='tomato', label='Reference')
    ax.fill_between(vchannels_fold, data[:, ipix, jpix], step='mid', facecolor='tomato', alpha=0.1)

    ax.step(vchannels_fold, data[:, imir, jmir], where='mid', lw=3.3, color='dodgerblue', label='Mirror')
    ax.fill_between(vchannels_fold, data[:, imir, jmir], step='mid', facecolor='dodgerblue', alpha=0.1)

    ax.step(-vchannels_fold, data[:, imir, jmir], where='mid', lw=2.2, color='k', label='Folded mirror')
    ax.fill_between(-vchannels_fold, data[:, imir, jmir], step='mid', facecolor='0.5', alpha=0.6, zorder=20)
    ax.fill_between(-vchannels_fold, data[:, imir, jmir], step='mid', facecolor='none', edgecolor='k', lw=0, hatch='X', zorder=21)

ax.axvline(vchannels[args.channel_id]-vsys, lw=2.0, dash_capstyle='round', dashes=(2.0, 2.5), color='magenta', label=r'$\upsilon_{\rm ref}$')
make_up_ax(ax, labelbottom=True, labeltop=False, labelsize=MEDIUM_SIZE+1, xlims=(-4.5, 4.5), ylims=(-2, 1.2*np.nanmax(data[:, ipix, jpix])))
make_1d_legend(ax, fontsize=MEDIUM_SIZE+2, handlelength=1.5, ncols=4)
mod_major_ticks(ax, axis='x', nbins=10)

if args.units=='Jy':
    ax.set_ylabel('Intensity [mJy/beam]', fontsize=MEDIUM_SIZE+2)
else:
    ax.set_ylabel('Brightness temperature [K]', fontsize=MEDIUM_SIZE+2)
ax.set_xlabel(r'Velocity relative to CSD $\upsilon_{\rm sys}$ [km/s]', fontsize=MEDIUM_SIZE+2)

plt.savefig('spectra_reference_and_mirror.png', bbox_inches='tight', dpi=200)
show_output(args)


#**************
#FOLD AND PLOT
#**************
fig, ax, residuals_spec = fold_and_plot_spectra(datacube, ijlist, vsys=vsys)

ax.axvline(vchannels[args.channel_id] - vsys, ls='--', lw=2.0, dash_capstyle='round', dashes=(2.0, 2.5), color='magenta')

plt.savefig('spectra_folded_residuals_%dpix.png'%(args.dpix), bbox_inches='tight', dpi=200)
show_output(args)


#*****************************
#REFERENCE CHANNEL AND PIXELS
#*****************************
def plot_projected_channel(chanid, tag=''):
    fig, ax = make_round_map(moment_data, levels_im, Xproj*u.m, Yproj*u.m, args.Router*Rout*u.au,
                         z_func=z_func, z_pars=z_pars, incl=incl, PA=PA, xc=xc, yc=yc,
                             cmap=cmap_mom, clabel=unit, fmt=cfmt, 
                             gaps=gaps, rings=rings, kinks=kinks,
                             make_contourf=False,
                             make_contour=False,
                             make_radial_grid=False,
                             make_azimuthal_grid=True,                             
                             make_cbar=False,
                             #ticks_phi=[],                             
                             #mask_inner=args.Rinner*datacube.beam_size
    )
    
    cf_map = ax.contourf(X, Y, datacube.data[chanid], levels=levels_im, cmap=cmap_mom)
        
    for ij in ijlist:
        i,j = ij
        ax.scatter(X[i,j], Y[i,j], marker='o', fc='none', ec='tomato', lw=1.5, s=100)

    ax.scatter(Rp_au*np.cos(phip_rad), Rp_au*np.sin(phip_rad), marker='X', fc='none', ec='k', lw=2.5, s=350)
    ax.scatter(X[imir,jmir], Y[imir,jmir], marker='P', fc='none', ec='k', lw=2.5, s=350)

    ax.set_title('Projected channel map (%.1f km/s)'%vchannels[chanid], fontsize=MEDIUM_SIZE+2)
    ax.set_xlim(-Rout, Rout)
    ax.set_ylim(-Rout, Rout)

    plt.savefig('spectra_mappixels_%dpix_%s.png'%(args.dpix, tag), bbox_inches='tight', dpi=200)    
    show_output(args)

plot_projected_channel(args.channel_id, tag='reference')
plot_projected_channel(get_mirror_channel_index(vchannels, args.channel_id, vsys), tag='mirror')




from discminer.mining_control import _mining_intensdistrib
from discminer.core import Data

from discminer.mining_utils import (_get_mask_tuples,
                                    _merge_R_phi_mask,
                                    make_masks,
                                    get_noise_mask,
                                    load_disc_grid,
                                    get_2d_plot_decorators,
                                    init_data_and_model,
                                    load_moments,
                                    mark_planet_location,
                                    show_output)

from discminer.plottools import (make_up_ax,
                                 make_round_map,
                                 use_discminer_style)

import json
import random
import numpy as np
from astropy import units as u
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde, skew, kurtosis

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_intensdistrib(None)
    args = parser.parse_args()

#*******************
#FILL UP EMPTY ARGS
#*******************
nmasks = int(round(0.5*len(args.annuli)))

if len(args.r0)==nmasks:
    rcenters = args.r0    
else:
    nmissing = nmasks - len(args.r0)
    rcenters = np.append(args.r0, [0]*nmissing)
    
if len(args.phi0)==nmasks:
    phicenters = np.radians(args.phi0)
else:
    nmissing = nmasks - len(args.phi0)
    phicenters = np.radians(np.append(args.phi0, [0]*nmissing))

if len(args.colors)==nmasks:
    colors = args.colors
else:
    nmissing = nmasks - len(args.colors)
    colors = np.append(args.colors, ['black']*nmissing)

if len(args.linewidths)==nmasks:
    lws = args.linewidths
else:
    nmissing = nmasks - len(args.linewidths)
    lws = np.append(args.linewidths, [1.5]*nmissing)
    
#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

vel_sign = best['velocity']['vel_sign']
vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']
PA = best['orientation']['PA']
xc = best['orientation']['xc']
yc = best['orientation']['yc']

gaps = custom['gaps']
rings = custom['rings']
try:
    kinks = custom['kinks']
except KeyError:
    kinks = []

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment)

if args.clim!=0:
    clim = args.clim
    
#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']
au_to_m = u.au.to('m')

dpc = meta['dpc']*u.pc

R_lev = np.arange(25, 0.98*Rout, 50)*au_to_m

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model(Rmin=0, Rmax=1.6)
noise_mean, mask = get_noise_mask(datacube, thres=3)
modelcube = Data('cube_model_%s.fits'%tag, dpc)

datapdf = np.where(mask[None, :, :], np.nan, datacube.data)

#Definitions for plots
beam_au = datacube.beam_size.to('au').value
if args.absolute_Rinner>=0:
    Rmod_in = args.absolute_Rinner
else:
    Rmod_in = args.Rinner*beam_au

if args.absolute_Router>=0:
    Rmod_out = args.absolute_Router
else:
    Rmod_out = args.Router*Rout

xmax = model.skygrid['xmax'] 
xlim = 1.1*Rout
extent= np.array([-xmax, xmax, -xmax, xmax])/au_to_m
#plot_title = ctitle + r' $-$ ' + clabel.split('[')[0]

#****************************
#LOAD DISC GEOMETRY AND MASK
#****************************
R, phi, z = load_disc_grid()

R_s = R[args.surface]*u.m.to('au')
phi_s = np.degrees(phi[args.surface])

Xproj = R_s*np.cos(phi[args.surface])
Yproj = R_s*np.sin(phi[args.surface])

#*************************
#LOAD AND CLIP MOMENT MAPS
#Clip radially
# Masked sections are visually drawn at the end
moment_data, moment_model, residuals, mtags = load_moments(
    args,
    mask=mask, #fill masked cells with nans
    clip_Rmin=0.0*u.au,
    clip_Rmax=Rmod_out*u.au,
    clip_Rgrid=R_s*u.au
)

#***********
#MAKE MASKS
#***********
masktuples_R = _get_mask_tuples(args.annuli)
masktuples_phi = _get_mask_tuples(args.wedges)

if len(masktuples_phi)==nmasks:
    pass
else:
    nmissing = nmasks - len(masktuples_phi)
    masktuples_phi = masktuples_phi + [[]]*nmissing
    
masks = []
xi, yi = [], []

for i in range(nmasks):

    if rcenters[i] != 0:
        x0 = rcenters[i]*np.cos(phicenters[i])
        y0 = rcenters[i]*np.sin(phicenters[i])
        Rgrid = np.hypot(Xproj - x0, Yproj-y0)
        phigrid = np.arctan2(Yproj - y0, Xproj - x0)
        xi.append(x0)
        yi.append(y0)

    else:
        Rgrid = R_s
        phigrid = phi_s
        xi.append(0)
        yi.append(0)

    mask_R={'map2d': Rgrid, 'lims': masktuples_R[i]}        
    mask_phi={'map2d': phigrid, 'lims': masktuples_phi[i]}

    mask0 = _merge_R_phi_mask(mask_R, mask_phi)
    if i>0:
        maski = mask0 & (~masks[i-1]) 
    else:
        maski = mask0
        
    masks.append(maski)

#*********
#MAKE PDF
#*********
chanwidth = np.abs(np.median(np.diff(datacube.vchannels)))
dv_native = 0.5*chanwidth

vmin = -5*dv_native + np.min(datacube.vchannels)   # generous margins
vmax =  5*dv_native + np.max(datacube.vchannels)
bins   = np.arange(vmin, vmax + dv_native, dv_native)
vcenters = 0.5*(bins[:-1] + bins[1:])

def thick_gaussian(v, sigma, tau=0.0, noise_mean=0.05, seed=10): #SNR 20 at planet location
    """
    Compute a thick (flat-topped) Gaussian intensity profile,
    optionally adding random Gaussian noise.

    Parameters
    ----------
    v : array-like
        Velocity or x-grid (centered at zero recommended).
    sigma : float
        Standard deviation of the underlying Gaussian (same units as v).
    tau : float, optional
        Optical depth factor. tau=0 gives a normal Gaussian,
        larger tau flattens the peak (optically thick limit).
    noise_mean : float, optional
        RMS of additive Gaussian noise (default 0 = no noise).
    rng : np.random.Generator, optional
        Random number generator for reproducibility (e.g., np.random.default_rng(42)).

    Returns
    -------
    I : ndarray
        Normalized intensity profile (max=1 if noise_mean=0).
    """
    G = np.exp(-0.5 * (v / sigma)**2)

    if tau == 0.0:
        I = G.copy()
    else:
        # Radiative transfer form: (1 - exp(-tau(v))) normalized to the line center
        I = (1.0 - np.exp(-tau * G)) / (1.0 - np.exp(-tau))

    # Add Gaussian-distributed random noise
    if noise_mean > 0.0:
        if seed is None:
            rng = np.random.default_rng()
        else:
            rng = np.random.default_rng(seed)
            
        I += rng.normal(0.0, noise_mean, size=I.shape)

    # Clip to non-negative and renormalize if no saturation wanted
    I = np.clip(I, 0.0, None)
    if I.max() > 0:
        I /= I.max()

    return I


def thick_gaussian_random(vmin, vmax, sigma, tau=0.0, noise_mean=0.05, nsamples=50000,
                          seed=10, normalize_after_noise=True):
    """
    Generate a thick (flat-topped) Gaussian intensity distribution
    by sampling random velocities between vmin and vmax.

    Parameters
    ----------
    vmin, vmax : float
        Velocity range over which to draw random samples.
    sigma : float
        Standard deviation of the underlying Gaussian (same units as vmin/vmax).
    tau : float, optional
        Optical-depth-like factor. tau=0 -> pure Gaussian.
    noise_mean : float, optional
        RMS of additive Gaussian noise (0 = no noise).
    nsamples : int, optional
        Number of random velocity samples to draw.
    seed : int or None
        RNG seed for reproducibility.
    normalize_after_noise : bool, optional
        If True, renormalize max(I)=1 after adding noise.
        If False, keep the pre-noise peak normalized to 1.

    Returns
    -------
    v : ndarray
        Random velocity samples (uniformly distributed between vmin and vmax).
    I : ndarray
        Corresponding intensity samples, normalized to max=1.
    """
    rng = np.random.default_rng(seed)
    v = rng.uniform(vmin, vmax, nsamples)

    G = np.exp(-0.5 * (v / sigma)**2)

    if tau == 0.0:
        I = G
    else:
        # Use expm1 for numerical stability
        num = -np.expm1(-tau * G)
        den = -np.expm1(-tau)
        I = num / den

    # Add Gaussian-distributed noise
    if noise_mean > 0.0:
        I_noisy = I + rng.normal(0.0, noise_mean, size=I.shape)
        I_noisy = np.clip(I_noisy, 0.0, None)
        if normalize_after_noise and I_noisy.max() > 0:
            I_noisy /= I_noisy.max()
        I = I_noisy
    else:
        if I.max() > 0:
            I /= I.max()

    return v, I

def gaussian_reference_pdf(sigma, tau):
    vmin = -4*sigma
    vmax = 4*sigma
    v = np.arange(vmin, vmax+chanwidth, chanwidth)
    I = thick_gaussian(v, sigma, tau)
    I = I[np.abs(v)<4*sigma]

    #v, I = thick_gaussian_random(vmin, vmax, sigma, tau=tau)

    kde = gaussian_kde(I)
    Igrid = np.linspace(I.min(), I.max(), 300)
    pdf = kde(Igrid)
    return Igrid, pdf

def bin_add(v_dep, y, perbin):
    # v_dep: deprojected velocity axis for a spectrum (datacube.vchannels - vcent)
    # y: normalized intensities
    idx = np.digitize(v_dep, bins) - 1   # bin indices
    good = (idx >= 0) & (idx < len(vcenters)) & np.isfinite(y)
    for k in np.where(good)[0]:
        perbin[idx[k]].append(y[k])
        
def make_pdf(maski):
    intensities = datapdf[:, maski].ravel()
    intensities = intensities[intensities>args.sigma*noise_mean]
    intensities = intensities/np.nanmax(intensities)
    kde = gaussian_kde(intensities) #Normalise to peak intensity
    Igrid = np.linspace(np.nanmin(intensities), np.nanmax(intensities), 300)
    pdf = kde(Igrid)
    return Igrid, pdf

def make_spectra(maski):
    intensities = datapdf[:, maski]
    return intensities.T

def select_spectra(spectra, n):
    inds = random.sample(range(0, len(spectra)), n)
    return spectra[inds]

#**************
#MAKE PLT AXES
#**************
figx, figy = 12, 4
x2y = figx/figy

fig = plt.figure(figsize=(figx,figy))

axr = fig.add_axes([0.05, 0.05, 0.9/x2y, 0.9])
axp0 = 0.05+0.9/x2y+0.03
axp = fig.add_axes([axp0, 0.14, 0.90-axp0, 0.72])

if len(args.spectra)>0:
    axs = fig.add_axes([0.94, 0.22, 0.4, 0.56])
    spectra, draws = [], []
    zup, vphi, vcent = [], [], []

sigma_ref = 4 * dv_native #0.3 km/s
tau_ref = 20
Igrid_ref, pdf_ref = gaussian_reference_pdf(sigma_ref, tau_ref)
axp.plot(Igrid_ref, pdf_ref, ls='--', color='k', lw=1.0,
         label=fr'Gaussian ref ($\sigma={sigma_ref:.2f}$ km/s)', zorder=5)
print ('Thick-Gaussian skewness:', skew(pdf_ref))

Igrid_ref, pdf_ref = gaussian_reference_pdf(sigma_ref, tau_ref*0)
axp.plot(Igrid_ref, pdf_ref, ls=':', color='k', lw=1.0,
         label=fr'Gaussian ref ($\sigma={sigma_ref:.2f}$ km/s)', zorder=5)
print ('Gaussian skewness:', skew(pdf_ref))

#Ig = thick_gaussian(vcenters, sigma_ref, tau_ref*0)
#axs.plot(vcenters, Ig, 'tomato', lw=5, zorder=100)

ns = 0 #spectra counter
for i in range(nmasks):
    Igrid, pdf = make_pdf(masks[i])
    axp.plot(Igrid, pdf, color=colors[i], lw=lws[i], zorder=50-i)
    axp.fill_between(Igrid, pdf, color='k', lw=0, alpha=0.05)    

    #skewi = skew(Igrid*pdf)
    #kurti = kurtosis(Igrid*pdf)

    skewi = skew(pdf)
    kurti = kurtosis(pdf)
    
    """
    nsamples = 5000
    samples = np.random.choice(Igrid, size=nsamples, replace=True, p=pdf/np.sum(pdf))

    # Compute skewness and kurtosis of the intensity distribution
    skew_samples = skew(samples)
    kurt_samples = kurtosis(samples)

    print(f"Mask {i}: skew(PDF)={skew(pdf):.3f}, kurt(PDF)={kurtosis(pdf):.3f}, "
          f"skew(samples)={skew_samples:.3f}, kurt(samples)={kurt_samples:.3f}")
    """
    #print ('PDF i skew kurt:', i, skew(pdf), kurtosis(pdf))

    if len(args.spectra)>0:
        
        if i in args.spectra:
            
            perbin = [[] for _ in range(len(vcenters))]

            spectrai = make_spectra(masks[i])
            spectra.append(spectrai)
            
            if len(spectrai) > args.ndraws:
                drawsi = random.sample(range(0, len(spectrai)), args.ndraws)
            else:
                drawsi = range(0, len(spectrai))

            draws.append(drawsi)
            zupi = model.z_upper_func({'R': Rgrid[masks[i]]*au_to_m}, **best['height_upper'])
            vphii = model.velocity_func({'R': Rgrid[masks[i]]*au_to_m, 'z': zupi}, **best['velocity'])
            vcenti = vsys + vphii * np.sin(incl) * np.cos(np.radians(phigrid[masks[i]]))            
            zup.append(zupi)
            vphi.append(vphii)
            vcent.append(vcenti)
            peaki = np.nanmax(spectrai[drawsi]) #Global peak

            skewall = []
            kurtall = []            
            for j in drawsi:
                #peaki = np.nanmax(spectrai[j]) #Peak per spectrum
                spec = spectrai[j]/peaki
                v_dep = datacube.vchannels-vcenti[j]                
                bin_add(v_dep, spec, perbin)
                axs.plot(v_dep, spec, lw=0.1, color=colors[i], alpha=0.2, zorder=50-i)

                skewall.append(skew(spec))
                kurtall.append(kurtosis(spec))

            print ('Spectra i med(skew), med(kurt)', i, np.median(skewall), np.median(kurtall))
            
            med_i = np.empty(len(vcenters)); med_i[:] = np.nan
            p16_i = np.empty(len(vcenters)); p16_i[:] = np.nan
            p84_i = np.empty(len(vcenters)); p84_i[:] = np.nan
            cnt_i = np.zeros(len(vcenters), dtype=int)

            for b in range(len(vcenters)):
                if perbin[b]:
                    arr = np.asarray(perbin[b], float)
                    cnt_i[b] = arr.size
                    med_i[b] = np.median(arr)
                    p16_i[b] = np.percentile(arr, 16)
                    p84_i[b] = np.percentile(arr, 84)

            # (optional) mask bins with too few samples
            min_count = 3
            sparse = cnt_i < min_count
            med_i[sparse] = np.nan; p16_i[sparse] = np.nan; p84_i[sparse] = np.nan

            # ----- plot per-i median + spread -----
            if colors[i] in ['k', 'black']:
                fci = '0.9'
            else:
                fci = colors[i]
                
            axs.scatter(vcenters, med_i, lw=1.5, ec='k', color=fci, s=30, zorder=100-i)
            
            #axs.plot(vcenters, med_i, lw=4.0, color=colors[i], alpha=0.4, zorder=99-i)        
            #axs.fill_between(vcenters, p16_i, p84_i, alpha=0.15, color=colors[i], linewidth=0, zorder=90-i)

            #axs.text(0.20, 0.85-0.12*ns, '%.2f'%skewi, fontsize=12, ha='right', va='top', color=colors[i], transform=axs.transAxes)
            ns += 1        

            """
            good = np.isfinite(med_i)
            v = vcenters[good]
            I = med_i[good]

            if I.size == 0 or np.all(I <= 0):
                print(f"[warn] Empty or zero med_i for region {i}")
            else:
                I = np.clip(I, 0, None)
                w_sum = np.sum(I)
                mu_v = np.sum(I * v) / w_sum
                var_v = np.sum(I * (v - mu_v)**2) / w_sum
                std_v = np.sqrt(var_v)
                mu3_v = np.sum(I * (v - mu_v)**3) / w_sum
                skew_weighted_velocity = mu3_v / (std_v**3)
                
            print(f"Region {i}: intensity-weighted velocity skewness = {skew_weighted_velocity:.3f}")
            """
    axp.text(0.2, 0.85-0.09*i, '%.2f'%skewi, fontsize=args.fontsize-1, ha='right', va='top', color=colors[i], transform=axp.transAxes)

axp.text(0.2, 0.92, 'Skewness', weight='bold', fontsize=args.fontsize-1, ha='right', va='top', color='0.6', transform=axp.transAxes)

make_up_ax(axp, labelleft=False, labelright=True, labeltop=False, labelbottom=True, labelsize=16)
axp.set_xticks([0.0, 0.5, 1.0])
axp.set_yticks([0, 1, 2])
axp.set_xticklabels([0, r'0.5I$_{\rm peak}$', r'I$_{\rm peak}$'])

#axp.text(0, 1.02, 'Intensity Distribution of Velocity Channels', fontsize=12, transform=axp.transAxes)
if args.show_title:
    axp.text(0.5, 1.06, 'Probability Density of Intensities across Velocity Channels', fontsize=args.fontsize-2.5, transform=axp.transAxes, ha='center', va='center')

axp.yaxis.set_label_position("right")
#axp.set_ylabel('Probability density', fontsize=14, labelpad=15)
axp.set_xlim(0, 1)
axp.set_ylim(0, None)

if len(args.spectra)>0:    
    make_up_ax(axs, labelleft=False, labelright=True, labeltop=False, labelbottom=True, labelsize=14)
    axs.tick_params(axis='y', pad=-80, labelcolor='k')
    axs.axvline(0, dash_capstyle='round', dashes=(3.0, 2.5), color='k', zorder=0)
    axs.axhline(0.5, xmin=0.5, dash_capstyle='round', dashes=(0.5, 1.5), color='k', zorder=0)
    axs.axhline(1.0, xmin=0.5, dash_capstyle='round', dashes=(0.5, 1.5), color='k', zorder=0)
    axs.set_yticks([0.5, 1.0])
    axs.set_yticklabels([r'0.5I$_{\rm peak}$', r'I$_{\rm peak}$'])

    #axs.text(0.20, 0.95, 'Skewness', fontsize=12, ha='right', va='top', color='0.6', transform=axs.transAxes)

    bbox = dict(boxstyle="round", ec="w", fc="w", alpha=1)
    plt.setp(axs.get_yticklabels(), bbox=bbox)

    axs.set_xticks([-0.5, 0.0, 0.5])
    axs.set_xlabel('Stacked velocities [km/s]')
    axs.set_xlim(-1.1, 1.1)
    axs.set_ylim(0, 1.07)

#***************
#MAKE ROUND MAP
#***************
kwargs_im = dict(cmap=cmap_mom, extent=extent, levels=levels_im)

if args.surface in ['up', 'upper']:
    z_func = model.z_upper_func
    z_pars = best['height_upper']

elif args.surface in ['low', 'lower']:
    z_func = model.z_lower_func
    z_pars = best['height_lower']

levels_resid = np.linspace(-clim, clim, 32)
        
fig, axr = make_round_map(1000*residuals, levels_resid, Xproj*u.au, Yproj*u.au, Rmod_out*u.au,
                          fig=fig, ax=axr,
                          kwargs_contourf=dict(cmap=cmap_res), #Skip set_under, set_over by overwriting cmap
                          z_func=z_func, z_pars=z_pars, incl=incl, PA=PA, xc=xc, yc=yc,
                          cmap=cmap_res, clabel='m/s', fmt='%d',
                          gaps=gaps, rings=rings, kinks=kinks,
                          make_cbar=args.colorbar,                         
                          mask_inner=Rmod_in*u.au,
                          fontsize_azimuthal_grid=args.fontsize-4,
                          fontsize_radial_grid=args.fontsize-3, 
                          fontsize_cbar=args.fontsize-2,
                          fontsize_xaxis=args.fontsize-4,
                          fontsize_nskyaxis=args.fontsize-2,
                          make_radial_lines=False,
                          make_nskyaxis=args.show_nsky,
                          make_Rout_proj=args.show_xaxis,
                          make_xaxis=args.show_xaxis,                          
)

#Filled regions
for i in np.arange(0, len(masks))[::-1]:
    
    if rcenters[i] != 0:
        colorcirc = 'k'
        alpha = 0.5
        lw = 2
    else:
        colorcirc = colors[i]
        alpha = 0.3
        lw = 1

    if np.any(masks[i]):
        axr.contourf(
            Xproj, Yproj, masks[i].astype(float),
            levels=[0.5, 1.5],
            colors=[colors[i]],
            alpha=alpha,
            #hatches=['///'],
            #zorder=30 - i,
            antialiased=True
        )

    circle = plt.Circle((xi[i], yi[i]), masktuples_R[i][-1], color=colorcirc, lw=lw, fill=False)#, zorder=10-i)
    axr.add_patch(circle)
    circle = plt.Circle((xi[i], yi[i]), masktuples_R[i][0], color=colorcirc, lw=lw, fill=False)#, zorder=10-i)
    axr.add_patch(circle)

if len(args.mask_R)>0 or len(args.mask_phi)>0:
    make_masks(axr, args.mask_R, args.mask_phi, Rmax=Rmod_out)
    
mark_planet_location(axr, args, edgecolors='k', lw=0, s=550, coords='disc', model=model)    

plt.savefig('intensity_distribution_%s_%s.png'%(meta['disc'], meta['mol']), bbox_inches='tight', dpi=200)    
show_output(args)

